/**
   tdslib.c
   k.furukawa, mar.19.2003.

   k.f, mar.20.2003.  check return value
   k.f, mar.30.2003.  use strlen instead of sizeof
   k.f, mar.30.2003.  shorter timeout (this did not work)
   k.f, mar.30.2003.  intercept signal to clean up communication
   k.f, apr.5.2003.   separate library routines (tdstest4 v1.6)
   k.f, apr.6.2003.   allocate structure locally (v1.6.1)
   k.f, apr.7.2003.   return string (v1.6.2)
   k.f, apr.10.2003.  library for tcltds (tdslib v1.7)
   k.f, apr.11.2003.  use tcltds.h (tdslib v1.7.1)
   k.f, sep.5.2003.   use clnt_control(timeout), device_lock/unlock (v1.8)
   k.f, sep.6.2003.   check magic, add tds_timeout
   k.f, dec.10.2003.  remove device_abort from tds_open (v1.8.5)
   k.f, sep.3.2007.   ignore error at device_async for dso5054a (v1.9.0)
   k.f, sep.4.2007.   'gpib0' for tektronix, 'inst0' for agilent (v1.9.1)
   k.f, dec.9.2007.   larger buffer size (v1.9.2)
   k.f, aug.27.2008.  made 'inst0' the default (v2.0)

**/

static char     what_tdslib[] =
"@(#)tdslib.c v2.0.0 Tds/Vxi library, K.Furukawa, Mar.2003-Aug.2008";

/**
  This file defines 3 local functions and 7 global functions

   Local functions:
   tds_malloc  --  wrapper for malloc
   tds_free  --  wrapper for free
   tds_signalhandle  --  possible signal handler (does not work ?)

   Global functions:
   int tds_init( tds_struct **to_st, char *to_devip )
     --  initialize local rpc data structures
   int tds_open( tds_struct *to_st )
     --  open rpc connection to a tds device
   int tds_timeout( tds_struct *to_st, int millisec )
     --  specify processing time-out in milli-seconds
   int tds_write( tds_struct *to_st, char *to_writestring )
     --  write to a tds device
   int tds_read( tds_struct *to_st, char *to_print )
     --  read from a tds device
   int tds_lock( tds_struct *to_st )
     --  lock a tds device
   int tds_unlock( tds_struct *to_st )
     --  unlock a tds device
   int tds_close( tds_struct *to_st )
     --  close rpc connection

  This library works at least with following devices.
     Tektronix TDS3xxxB
     Tektronix DPO3xxx
     Tektronix DPO4xxx
     Tektronix DPO7xxx
     Tektronix DSA8xxx
     Agilent   DSO5xxx
**/
/*
  This file also defines a main routine for a test (see the symbol MAIN)
 */

#include <stdio.h>
#include <stdlib.h>
#include <rpc/rpc.h>
#include <signal.h>
#include <sys/types.h>
#include <string.h>

#include "vxi11.h"
#include "tdslib.h"

int tds_debug = 0;		/* debug mode */
int tds_gpib0 = 0;		/* old device name (gpib0) instead of (inst0) */
		/* old tektronix accept both gpib0 and inst0 */
		/* new tektronix and agilent accept only inst0 */

static tds_struct *last_to_st;

/*
 *  tds_malloc  --  wrapper for malloc
 */
static void *tds_malloc( size_t size )
{
  return (void *)malloc( size );		/* normal system malloc */
}

/*
 *  tds_free  --  wrapper for free
 */
static void tds_free( void *pointer )
{
  free( pointer );				/* normal system free */
  return;
}

/*
 *  tds_signalhandle  --  possible signal handler (does not work ?)
 */
static void tds_signalhandle( int signo )
{
  /* abort last command (really need?)*/
  if (last_to_st->to_rpc_abort != NULL) {
    last_to_st->dev_err = device_abort_1(&last_to_st->dev_link_resp->lid,
					 last_to_st->to_rpc_abort);
    if (last_to_st->dev_err == NULL) {
      clnt_perror(last_to_st->to_rpc_abort, last_to_st->to_dev_ip);
      exit(2);
    }
    if (last_to_st->dev_err->error != 0) {
      fprintf (stderr, "tds-signal -- Abort Error, Code = %d\n", 
	       last_to_st->dev_err->error);
      exit(2);
    }
    if (tds_debug) fprintf(stderr, "tds_signal -- device_abort\n");
  }

  /* close link */
  last_to_st->dev_err = destroy_link_1(&last_to_st->dev_link_resp->lid,
				       last_to_st->to_rpc_core);
  if (last_to_st->dev_err == NULL) {
    clnt_perror(last_to_st->to_rpc_core, last_to_st->to_dev_ip);
    exit(2);
  }
  if (tds_debug) fprintf(stderr, "tds_signal -- destroy_link\n");
  
  exit(1);
}

/**
 *  int tds_init( tds_struct **to_st, char *to_devip )
 *    --  initialize local rpc data structures
**/
int tds_init(tds_struct **to_st, char *to_devip)
{
  tds_struct *new_to_st;

  new_to_st = tds_malloc( sizeof (tds_struct) );
  if (new_to_st == NULL) {
      perror("tds_malloc");
      return -1;
  }
  strncpy(new_to_st->to_dev_ip, to_devip, TDS_DEVIPMAX);
  new_to_st->magic = TDS_MAGIC;
  new_to_st->timeout = TDS_TIMEOUT;
  
  /* connect RPC server(core link) */
  new_to_st->to_rpc_core 
    = clnt_create (new_to_st->to_dev_ip, DEVICE_CORE, 
		   DEVICE_CORE_VERSION, "tcp");
  if (new_to_st->to_rpc_core == NULL) {
    clnt_pcreateerror(new_to_st->to_dev_ip);
    return -1;
  }
  if (tds_debug) fprintf(stderr, "tds_init -- clnt_create DEVICE_CORE\n");

  {
    struct timeval timeout = { TDS_TIMEOUT/1000, (TDS_TIMEOUT%1000)*1000 };

    clnt_control(new_to_st->to_rpc_core, CLSET_TIMEOUT, (char *) &timeout);
    if (tds_debug) fprintf(stderr, "tds_init -- clnt_control TIMEOUT\n");
  }
  
  /* connect RPC server(abort link) */
  new_to_st->to_rpc_abort 
    = clnt_create (new_to_st->to_dev_ip, DEVICE_ASYNC, 
		   DEVICE_ASYNC_VERSION, "tcp");
  if (new_to_st->to_rpc_abort == NULL) {
    clnt_pcreateerror(new_to_st->to_dev_ip);
    if (tds_debug) 
	fprintf(stderr, "tds_init -- clnt_create DEVICE_ASYNC failed (ignored)\n");
    /* return -1; */		/* dso5xxx does not support DEVICE_ASYNC */
  } else {
    if (tds_debug) fprintf(stderr, "tds_init -- clnt_create DEVICE_ASYNC\n");
  }

  last_to_st = new_to_st;
  *to_st = new_to_st;

  return 0;
}

/**
 *  int tds_open( tds_struct *to_st )
 *    --  open rpc connection to a tds device
**/
int tds_open(tds_struct *to_st)
{
  Create_LinkParms dev_link_param;
  
  if (to_st->magic != TDS_MAGIC) {
    fprintf (stderr, "tds-open -- un-initialized or wrong structure\n");
    return -1;
  }

  /* setup VXI11 link param */
  dev_link_param.clientId = (long)to_st->to_rpc_core;
  dev_link_param.lockDevice = 0;
  dev_link_param.lock_timeout = to_st->timeout; /* milli sec */
  if (tds_gpib0) dev_link_param.device = "gpib0";
  else dev_link_param.device = "inst0";

  to_st->dev_link_resp = create_link_1(&dev_link_param, to_st->to_rpc_core);
  if (to_st->dev_link_resp == NULL) {
    clnt_perror(to_st->to_rpc_core, to_st->to_dev_ip);
    return -1; 
  }
  if (tds_debug) fprintf(stderr, "tds_open -- create_link\n");

  signal(SIGINT, tds_signalhandle);
  signal(SIGSEGV, tds_signalhandle);
  signal(SIGTERM, tds_signalhandle);
  if (tds_debug) fprintf(stderr, "tds_open -- signal\n");
  
#ifdef OPEN_NEEDS_ABORT /* this part is not needed i think */

  /* abort last command (really need?)*/
  if (to_st->to_rpc_abort != NULL) {
    to_st->dev_err = device_abort_1(&(to_st->dev_link_resp)->lid, to_st->to_rpc_abort);
    if (to_st->dev_err == NULL) {
      clnt_perror(to_st->to_rpc_abort, to_st->to_dev_ip);
      return -1;
    }
    if (to_st->dev_err->error != 0) {
      fprintf (stderr, "Abort Error, Code = %d\n", to_st->dev_err->error);
      return -1;
    }
    if (tds_debug) fprintf(stderr, "tds_open -- device_abort\n");
  }

#endif /* OPEN_NEEDS_ABORT */

  return 0;
}

/**
 *  int tds_timeout( tds_struct *to_st, int millisec )
 *    --  specify timeout in milli-seconds
**/
int tds_timeout(tds_struct *to_st, int millisec)
{
  Device_Error *dev_err;

  if (to_st->magic != TDS_MAGIC) {
    fprintf (stderr, "tds-timeout -- un-initialized or wrong structure\n");
    return -1;
  }

  /* set rpc timeout, 300 millisecond larger than vxi11 timeout */
  {
    struct timeval timeout;
    timeout.tv_sec = (millisec+300)/1000;
    timeout.tv_usec = ((millisec+300)%1000)*1000;

    clnt_control(to_st->to_rpc_core, CLSET_TIMEOUT, (char *) &timeout);
    if (tds_debug) fprintf(stderr, "tds_timeout -- clnt_control TIMEOUT\n");
  }

  to_st->timeout = millisec;
  
  if (tds_debug) fprintf(stderr, "tds_timeout -- specify timeout\n");

  return 0;
}

/**
 *  int tds_write( tds_struct *to_st, char *to_writestring )
 *    --  write to a tds device
**/
int tds_write(tds_struct *to_st, char *to_writestring)
{
  Device_WriteParms dev_write_param;
  Device_WriteResp *dev_write_resp;

  if (to_st->magic != TDS_MAGIC) {
    fprintf (stderr, "tds-write -- un-initialized or wrong structure\n");
    return -1;
  }

  /* setup VXI11 write param */
  dev_write_param.lid = to_st->dev_link_resp->lid;
  dev_write_param.io_timeout = to_st->timeout; /* milli sec */
  dev_write_param.lock_timeout = to_st->timeout;
  dev_write_param.flags = 8 ; /* set End bit */
  dev_write_param.data.data_len = strlen(to_writestring);
  dev_write_param.data.data_val = to_writestring;

  dev_write_resp = device_write_1(&dev_write_param, to_st->to_rpc_core);
  if (dev_write_resp == NULL) {
      clnt_perror(to_st->to_rpc_core, to_st->to_dev_ip);
      /* return 1; */
      return -1;
  }
  if (dev_write_resp->error != 0) {
    fprintf (stderr, "Write Error, Code = %d\n", dev_write_resp->error);
    return -1;
  }
  if (tds_debug) fprintf(stderr, "tds_write -- device_write\n");

  return 0;
}

/**
 *  int tds_read( tds_struct *to_st, char *to_print )
 *    --  read from a tds device
**/
int tds_read(tds_struct *to_st, char *to_print)
{
  Device_ReadParms dev_read_param;
  Device_ReadResp *dev_read_resp;
  
  if (to_st->magic != TDS_MAGIC) {
    fprintf (stderr, "tds-read -- un-initialized or wrong structure\n");
    return -1;
  }

  /* setup VXI11 read param */
  dev_read_param.lid = to_st->dev_link_resp->lid;
  dev_read_param.requestSize = 1000; /* read 1000 byte */
  dev_read_param.io_timeout = to_st->timeout;
  dev_read_param.lock_timeout = to_st->timeout;
  dev_read_param.flags = 0;
  dev_read_param.termChar = 0;

  do {
    int i;
    
    dev_read_resp = device_read_1(&dev_read_param, to_st->to_rpc_core);
    if (dev_read_resp == NULL) {
	clnt_perror(to_st->to_rpc_core, to_st->to_dev_ip);
	return -1;
    }
    if (dev_read_resp->error != 0) {
      fprintf (stderr, "Read Error, Code = %d\n", dev_read_resp->error);
      return -1;
    }
    memcpy(to_print, &(dev_read_resp->data.data_val[0]), 
	   dev_read_resp->data.data_len);
    to_print += dev_read_resp->data.data_len;
    
  } while (dev_read_resp->reason < 4);

  *to_print = NULL;
  
  if (tds_debug) fprintf(stderr, "tds_read -- device_read\n");

  return 0;
}

/**
 *  int tds_lock( tds_struct *to_st )
 *    --  lock a tds device
**/
int tds_lock(tds_struct *to_st)
{
  Device_LockParms dev_lock_param;
  Device_Error *dev_err;

  if (to_st->magic != TDS_MAGIC) {
    fprintf (stderr, "tds-lock -- un-initialized or wrong structure\n");
    return -1;
  }

  /* setup VXI11 lock param */
  dev_lock_param.lid = to_st->dev_link_resp->lid;
  /* dev_lock_param.lock_timeout = TIMEOUT; */
  dev_lock_param.lock_timeout = to_st->timeout;
  dev_lock_param.flags = 1 ; /* set WaitLock bit */

  dev_err = device_lock_1(&dev_lock_param, to_st->to_rpc_core);
  if (dev_err == NULL) {
      clnt_perror(to_st->to_rpc_core, to_st->to_dev_ip);
      /* return 1; */
      return -1;
  }
  if (dev_err->error != 0) {
    fprintf (stderr, "Lock Error, Code = %d\n", dev_err->error);
    return -1;
  }
  if (tds_debug) fprintf(stderr, "tds_lock -- device_lock\n");

  return 0;
}

/**
 *  int tds_unlock( tds_struct *to_st )
 *    --  unlock a tds device
**/
int tds_unlock(tds_struct *to_st)
{
  Device_Link dev_link;
  Device_Error *dev_err;

  if (to_st->magic != TDS_MAGIC) {
    fprintf (stderr, "tds-unlock -- un-initialized or wrong structure\n");
    return -1;
  }

  /* setup VXI11 unlock param */
  dev_link = to_st->dev_link_resp->lid;

  dev_err = device_unlock_1(&dev_link, to_st->to_rpc_core);
  if (dev_err == NULL) {
      clnt_perror(to_st->to_rpc_core, to_st->to_dev_ip);
      /* return 1; */
      return -1;
  }
  if (dev_err->error != 0) {
    fprintf (stderr, "Unlock Error, Code = %d\n", dev_err->error);
    return -1;
  }
  if (tds_debug) fprintf(stderr, "tds_unlock -- device_unlock\n");

  return 0;
}

/**
 *  int tds_close( tds_struct *to_st )
 *    --  close rpc connection
**/
int tds_close(tds_struct *to_st)
{
  if (to_st->magic != TDS_MAGIC) {
    fprintf (stderr, "tds-close -- un-initialized or wrong structure\n");
    return -1;
  }

  /* close link */
  to_st->dev_err = destroy_link_1(&to_st->dev_link_resp->lid,
				  to_st->to_rpc_core);
  if (to_st->dev_err == NULL) {
    clnt_perror(to_st->to_rpc_core, to_st->to_dev_ip);
    return -1;
  }
  tds_free( to_st );
  if (tds_debug) fprintf(stderr, "tds_close -- destroy_link\n");

  return 0;
}

#ifdef MAIN

/**
  Sample main routine

   Shell command usage:
      tdstest -- please specify arguments
      usage: tdstest {ip} {command} [{debug}]
      ex.:   tdstest 172.19.68.21 '*idn?'
      ex.:   tdstest 172.19.68.21 '*idn?' debug

**/

#define BUFSIZE 1024
static char toprint[BUFSIZE*1000];

int
main(int argc, char *argv[]) {

  char *write_string;
  char *dev_ip;			/* ip address */

#ifdef CHECK
#define check(str) printf("%d ",sizeof(str))
  check(CLIENT);
  check(Create_LinkParms);
  check(Create_LinkResp);
  check(Create_LinkResp);
  check(Device_Error);
  check(Device_WriteParms);
  check(Device_WriteResp);
  check(Device_ReadParms);
  check(Device_ReadResp);
  printf("\n");
  fflush(stdout);
#endif /* CHECK */

  if (argc < 3) {
      fprintf(stderr, "tdstest -- please specify arguments\n");
      fprintf(stderr, "%s\n", what_tdslib);
      fprintf(stderr, "usage: tdstest {ip} {command} [{debug},{gpib0}]\n");
      fprintf(stderr, "ex.:   tdstest 172.19.68.21 '*idn?'\n");
      fprintf(stderr, "ex.:   tdstest 172.19.68.21 '*idn?' debug\n");
      fprintf(stderr, "ex.:   tdstest 172.19.68.21 '*idn?' inst0 debug\n");
      fprintf(stderr, "ex.:   tdstest 172.19.68.21 '*idn?' gpib0 debug\n");
      fprintf(stderr, "ex.:   tdstest 172.19.68.21 ':ch1:scale 0.02;:ch1:scale?'\n");
      fprintf(stderr, "ex.:   tdstest 172.19.68.21 ':hor:scale 0.001;:hor:scale?'\n");
      exit(1);
  }
  
  dev_ip = argv[1];
  write_string = argv[2];
  {
    int i;
    for (i=3; i<argc; i++) {
      if (!strcmp("debug",argv[i])) tds_debug++;
      else if (!strcmp("agilent",argv[i])) tds_gpib0=0;
      else if (!strcmp("inst0",argv[i])) tds_gpib0=0;
      else if (!strcmp("tektronix",argv[i])) tds_gpib0++;
      else if (!strcmp("gpib0",argv[i])) tds_gpib0++;
      else fprintf(stderr, "tdstest -- illegal option '%s'\n", argv[i]);
    }
    if (tds_debug) 
      fprintf(stderr, "tdstest -- '%s', '%s', '%s'\n",
	      dev_ip, write_string, tds_gpib0?"gpib0":"inst0");
  }

  {
      /* int i;
      for (i=0; i<2; i++) */ {
	  tds_struct *tost;
    
	  int stat;
	  
	  stat = tds_init(&tost, dev_ip);
	  if (stat < 0) exit(-1);
	  
	  stat = tds_timeout(tost, 5000);
	  if (stat < 0) goto err_end;
	  
	  stat = tds_open(tost);
	  if (stat < 0) goto err_end;
	  
	  stat = tds_lock(tost);
	  if (stat < 0) goto err_end;
	  
	  stat = tds_write(tost, write_string);
	  if (stat < 0) goto err_end;
	  
	  stat = tds_read(tost, toprint);
	  if (stat < 0) goto err_end;

	  stat = tds_unlock(tost);
	  if (stat < 0) goto err_end;
	  
	  printf("%s", toprint);
	  
      err_end:
	  tds_close(tost);
      }
  }  
  exit(0);
}

#endif /* MAIN */
